  ' ipto_dos.bas                 July 8, 2017
  
  ' this MMBASIC program demonstrates how to optimize the
  ' two-body interplanetary orbit transfer problem using R. H. Gooding's
  ' numerical method described in the following RAE reports;
  
  ' R. H. Gooding, Technical Report 88027
  ' "On the Solution of Lambert's Orbital Boundary-Value Problem",
  ' Royal Aerospace Establishment, April 1988
  
  ' R. H. Gooding, Technical Memo SPACE 378
  ' "A Procedure for the Solution of Lambert's Orbital Boundary-Value Problem"
  ' Royal Aerospace Establishment, March 1991
  
  ' MMBASIC DOS version
  
  '''''''''''''''''''''
  
  option default float
  
  option base 1
  
  dim month$(12), pname$(8), jdleap(28), leapsec(28)
  
  dim dv1(3), dv2(3), vito(3), vfto(3), oevto(6), oevtf(6)
  
  dim ri(3), vi(3), rf(3), vf(3), oevi(6), oevf(6)
  
  ' number of control variables
  
  nv% = 2
  
  dim xg(nv%), fmin
  
  ' calendar months
  
  month$(1) = "January"
  month$(2) = "February"
  month$(3) = "March"
  month$(4) = "April"
  month$(5) = "May"
  month$(6) = "June"
  month$(7) = "July"
  month$(8) = "August"
  month$(9) = "September"
  month$(10) = "October"
  month$(11) = "November"
  month$(12) = "December"
  
  ' planet names
  
  pname$(1) = "Mercury"
  pname$(2) = "Venus"
  pname$(3) = "Earth"
  pname$(4) = "Mars"
  pname$(5) = "Jupiter"
  pname$(6) = "Saturn"
  pname$(7) = "Uranus"
  pname$(8) = "Neptune"
  
  ' Sun gravitational constant (km^3/sec^2)
  
  const xmu = 1.32712441933e11
  
  ' astronomical unit (kilometers)
  
  const aunit = 149597870.691
  
  ' utility constants
  
  const pi2 = 2.0 * pi, pidiv2 = 0.5 * pi
  
  const dtr = pi / 180.0, rtd = 180.0 / pi
  
  ' define "reference" julian date (1/1/2000)
  
  const jdate0 = 2451544.5
  
  print " "
  print "program ipto - interplanetary trajectory optimization"
  print "====================================================="
  
  ' minimize the sum of departure and arrival delta-v
  
  const otype% = 3
  
  ' Earth is the departure planet
  
  const ip1% = 3
  
  ' initial guess for departure calendar date and time
  
  cmonth1 = 6
  cday1 = 5
  cyear1 = 2003
  
  julian(cmonth1, cday1, cyear1, jdate1)
  
  ' Mars is the arrival planet
  
  const ip2% = 4
  
  ' initial guess for arrival calendar date and time
  
  cmonth2 = 12
  cday2 = 24
  cyear2 = 2003
  
  julian(cmonth2, cday2, cyear2, jdate2)
  
  ' "normalized" julian date initial guesses
  
  xg(1) = jdate1 - jdate0
  
  xg(2) = jdate2 - jdate0
  
  ' algorithm convergence criterion
  
  tol = 1.0e-4
  
  ' maximum number of algorithm iterations
  
  maxiter% = 200
  
  qnopt(nv%, tol, maxiter%, xg(), fmin, niter%, ier%, nsig%)
  
  ' compute optimal departure and arrival julian days
  
  jdopt1 = jdate0 + xg(1)
  
  jdopt2 = jdate0 + xg(2)
  
  ' evaluate optimal solution (solution passed via globals)
  
  usr_func (nv%, xg(), fmin)
  
  print " "
  print "initial orbital elements and state vector"
  print "(heliocentric ecliptic and equinox of J2000)"
  print "--------------------------------------------"
  print " "
  
  print pname$(ip1%)
  
  print " "
  
  jd2str(jdopt1)
  
  print " "
  
  planet(ip1%, jdopt1, oevi(), ri(), vi())
  
  oeprint(oevi())
  
  svprint(ri(), vi())
  
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''
  ' final heliocentric orbital elements and state vector
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''
  
  print " "
  print " "
  print "final orbital elements and state vector"
  print "(heliocentric ecliptic and equinox of J2000)"
  print "--------------------------------------------"
  print " "
  
  print pname$(ip2%)
  
  print " "
  
  jd2str(jdopt2)
  
  print " "
  
  planet(ip2%, jdopt2, oevf(), rf(), vf())
  
  oeprint(oevf())
  
  svprint(rf(), vf())
  
  print " "
  print "initial delta-v vector and magnitude"
  print "(heliocentric ecliptic and equinox of J2000)"
  print "--------------------------------------------"
  print " "
  
  print "delta-v vector x-component  ", str$(1000.0 * dv1(1), 0, 6) + " meters/second"
  print "delta-v vector y-component  ", str$(1000.0 * dv1(2), 0, 6) + " meters/second"
  print "delta-v vector z-component  ", str$(1000.0 * dv1(3), 0, 6) + " meters/second"
  
  print " "
  
  dvm1 = 1000.0 * vecmag(dv2())
  
  print "delta-v magnitude           ", str$(dvm1, 0, 6) + " meters/second"
  
  print " "
  print "orbital energy              ", str$(1.0e-6 * (dvm1 * dvm1), 0, 6) + " km^2/sec^2)"
  print " "
  
  print " "
  print "final delta-v vector and magnitude"
  print "(heliocentric ecliptic and equinox of J2000)"
  print "--------------------------------------------"
  print " "
  
  print "delta-v vector x-component  ", str$(1000.0 * dv2(1), 0, 6) + " meters/second"
  print "delta-v vector y-component  ", str$(1000.0 * dv2(2), 0, 6) + " meters/second"
  print "delta-v vector z-component  ", str$(1000.0 * dv2(3), 0, 6) + " meters/second"
  
  print " "
  
  dvm2 = 1000.0 * vecmag(dv2())
  
  print "delta-v magnitude           ", str$(dvm2, 0, 6) + " meters/second"
  
  print " "
  print "orbital energy              ", str$(1.0e-6 * (dvm2 * dvm2), 0, 6) + " km^2/sec^2)"
  print " "
  
  print " "
  print "total deltav magnitude      ", str$(dvm1 + dvm2, 0, 6) + " meters/second"
  print " "
  
  print "total orbital energy        ", str$(1.0e-6 * (dvm1 * dvm1 + dvm2 * dvm2), 0, 6) + " km^2/sec^2)"
  print " "
  
  print "total time of flight        ", str$(jdopt2 - jdopt1, 0, 6) + " days"
  
  print " "
  print " "
  print "transfer orbital elements and state vector after first delta-v"
  print "(heliocentric ecliptic and equinox of J2000)"
  print "--------------------------------------------"
  print " "
  
  eci2orb(xmu, ri(), vito(), oevto())
  
  oeprint(oevto())
  
  print " "
  print "transfer orbital elements and state vector prior to second delta-v"
  print "(heliocentric ecliptic and equinox of J2000)"
  print "--------------------------------------------"
  print " "
  
  eci2orb(xmu, rf(), vfto(), oevto())
  
  oeprint(oevto())
  
  print " "
  
end
  
  '''''''''''''''''''''''''''''''''''''''''''''''
  '''''''''''''''''''''''''''''''''''''''''''''''
  
sub glambert(cbmu, sv1(), sv2(), tof, vi(), vf())
  
  ' Gooding's solution of Lambert's problem
  
  ' input
  
  '  cbmu  = central body gravitational constant
  '  sv1() = initial 6-element state vector (position + velocity)
  '  sv2() = final 6-element state vector (position + velocity)
  '  tof   = time of flight (+ posigrade, - retrograde)
  
  ' output
  
  '  vi() = initial velocity vector of the transfer orbit
  '  vf() = final velocity vector of the transfer orbit
  
  ' References
  
  ' R. H. Gooding, Technical Report 88027
  ' On the Solution of Lambert's Orbital Boundary-Value Problem,
  ' Royal Aerospace Establishment, April 1988
  
  ' R. H. Gooding, Technical Memo SPACE 378
  ' A Procedure for the Solution of Lambert's Orbital Boundary-Value Problem
  ' Royal Aerospace Establishment, March 1991
  
  '''''''''''''''''''''''''''''''''''''''''''
  
  local ur1xv1(3), ux1(3), uy1(3), uz1(3), ux2(3)
  
  local urlxv1(3), uy2(3), uz2(3), vtmp1(3), vtmp2(3)
  
  local r1mag, r2mag, vr11, vr12, vt11, vt12
  
  r1mag = sqr(sv1(1) * sv1(1) + sv1(2) * sv1(2) + sv1(3) * sv1(3))
  
  r2mag = sqr(sv2(1) * sv2(1) + sv2(2) * sv2(2) + sv2(3) * sv2(3))
  
  for i% = 1 to 3
    
    vtmp1(i%) = sv1(i%)
    
    vtmp2(i%) = sv1(i% + 3)
    
  next i%
  
  vcross(vtmp1(), vtmp2(), ur1xv1())
  
  ur1xv1mag = vecmag(ur1xv1())
  
  for i% =1 to 3
    
    ur1xv1(i%) = ur1xv1(i%) / ur1xv1mag
    
  next i%
  
  for i% = 1 to 3
    
    ux1(i%) = sv1(i%) / r1mag
    
    ux2(i%) = sv2(i%) / r2mag
    
  next i%
  
  vcross(ux1(), ux2(), uz1())
  
  uz1mag = vecmag(uz1())
  
  for i% = 1 to 3
    
    uz1(i%) = uz1(i%) / uz1mag
    
  next i%
  
  ' calculate the minimum transfer angle
  
  theta = vdot(ux1(), ux2())
  
  if (theta > 1.0) then theta = 1.0
  
  if (theta < -1.0) then theta = -1.0
  
  theta = acos(theta)
  
  ' calculate the angle between the orbit normal of the
  ' initial orbit and the fundamental reference plane
  
  angle2on = vdot(ur1xv1(), uz1())
  
  if (angle2on > 1.0) then angle2on = 1.0
  
  if (angle2on < -1.0) then angle2on = -1.0
  
  angle2on = acos(angle2on)
  
  ' if angle to orbit normal is greater than 90 degrees
  ' and posigrade orbit, then flip the orbit normal
  ' and the transfer angle
  
  if ((angle2on > pidiv2) and (tof > 0.0)) then
    
    theta = pi2 - theta
    
    for i% = 1 to 3
      
      uz1(i%) = -uz1(i%)
      
    next i%
    
  end if
  
  for i% = 1 to 3
    
    uz2(i%) = uz1(i%)
    
  next i%
  
  vcross(uz1(), ux1(), uy1())
  
  uy1mag = vecmag(uy1())
  
  for i% = 1 to 3
    
    uy1(i%) = uy1(i%) / uy1mag
    
  next i%
  
  vcross(uz2(), ux2(), uy2())
  
  uy2mag = vecmag(uy2())
  
  for i% = 1 to 3
    
    uy2(i%) = uy2(i%) / uy2mag
    
  next i%
  
  vlamb(cbmu, r1mag, r2mag, theta, tof, vr11, vr12, vt11, vt12)
  
  for i% = 1 to 3
    
    vi(i%) = vr11 * ux1(i%) + vt11 * uy1(i%)
    
    vf(i%) = vr12 * ux2(i%) + vt12 * uy2(i%)
    
  next i%
  
end sub
  
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  
sub vlamb(cbmu, r1, r2, thr2, tdelt, vr11, vr12, vt11, vt12)
  
  ' Gooding lambert support subroutine
  
  ''''''''''''''''''''''''''''''''''''
  
  local dr, r1r2, r1r2th, csq, c, s, gms, qsqfm1, q
  
  local rho, sig, t, x1, qzminx, qzplx, zplqx
  
  ' calculate the chord of the triangle (c)
  
  thr2 = thr2 / 2.0
  
  dr = r1 - r2
  
  r1r2 = r1 * r2
  
  r1r2th = 4.0 * r1r2 * sin(thr2)^2
  
  csq = dr^2 + r1r2th
  
  c = sqr(csq)
  
  ' calculate the semiperimeter (s) of the triangle
  
  s = (r1 + r2 + c) / 2.0
  
  gms = sqr(cbmu * s / 2.0)
  
  qsqfm1 = c / s
  
  q = sqr(r1r2) * cos(thr2) / s
  
  ' check to see if the triangle is "sole case" (whole number of revolutions?)
  
  if (c <> 0.0) then
    
    rho = dr / c
    
    sig = r1r2th / csq
    
  else
    
    rho = 0.0
    
    sig = 1.0
    
  end if
  
  t = 4.0 * gms * tdelt / s^2
  
  xlamb(q, qsqfm1, t, x1)
  
  x = x1
  
  tlamb(q, qsqfm1, x, -1, qzminx, qzplx, zplqx)
  
  vt2 = gms * zplqx * sqr(sig)
  
  vr1 = gms * (qzminx - qzplx * rho) / r1
  
  vt1 = vt2 / r1
  
  vr2 = -gms * (qzminx + qzplx * rho) / r2
  
  vt2 = vt2 / r2
  
  vr11 = vr1
  
  vt11 = vt1
  
  vr12 = vr2
  
  vt12 = vt2
  
end sub
  
  ''''''''''''''''''''''''''''''''''''''''''
  ''''''''''''''''''''''''''''''''''''''''''
  
sub tlamb(q, qsqfm1, x, n%, dt, d2t, d3t, t)
  
  ' Gooding lambert support subroutine
  
  ''''''''''''''''''''''''''''''''''''
  
  const sw = 0.4
  
  local l1% = 0, l2% = 0, l3% = 0, lm1% = 0
  
  local qsq, xsq, z, u
  
  if (n% = -1) then lm1% = 1
  
  if (n% >= 1) then l1% = 1
  
  if (n% >= 2) then l2% = 1
  
  if (n% = 3) then l3% = 1
  
  qsq = q * q
  
  xsq = x * x
  
  u = (1.0 - x) * (1.0 + x)
  
  ' needed if series, and otherwise useful when z = 0
  
  if (lm1% = 0) then
    
    dt = 0.0
    
    d2t = 0.0
    
    d3t = 0.0
    
  end if
  
  if (lm1% = 1 or x < 0.0 or abs(u) > sw) then
    
    ' direct computation, not series
    
    y = sqr(abs(u))
    
    z = sqr(qsqfm1 + qsq * xsq)
    
    qx = q * x
    
    if (qx <= 0.0) then
      
      a = z - qx
      
      b = q * z - x
      
    end if
    
    if ((qx < 0.0) and (lm1% = 1)) then
      
      aa = qsqfm1 / a
      
      bb = qsqfm1 * (qsq * u - xsq) / b
      
    end if
    
    if ((qx = 0.0) and (lm1% = 1) or (qx > 0.0)) then
      
      aa = z + qx
      
      bb = q * z + x
      
    end if
    
    if (qx > 0.0) then
      
      a = qsqfm1 / aa
      
      b = qsqfm1 * (qsq * u - xsq) / bb
      
    end if
    
    if (lm1% = 1) then
      
      dt = b
      
      d2t = bb
      
      d3t = aa
      
    else
      
      if ((qx * u) >= 0.0) then
        
        g = x * z + q * u
        
      else
        
        g = (xsq - qsq * u) / (x * z - q * u)
        
      end if
      
      f = a * y
      
      if (x <= 1.0) then
        
        t = atan2(f, g)
        
      else
        
        if (f > sw) then
          
          t = log(f + g)
          
        else
          
          fg1 = f / (g + 1.0)
          
          term = 2.0 * fg1
          
          fg1sq = fg1 * fg1
          
          t = term
          
          twoi1 = 1.0
          
          told = 0.0
          
          do while (t <> told)
            
            twoi1 = twoi1 + 2.0
            
            term = term * fg1sq
            
            told = t
            
            t = t + term / twoi1
            
          loop
          
        end if
        
      end if
      
      t = 2.0 * (t / y + b) / u
      
      if (l1% = 1 and z <> 0.0) then
        
        qz = q / z
        
        qz2 = qz * qz
        
        qz = qz * qz2
        
        dt = (3.0 * x * t - 4.0 * (a + qx * qsqfm1) / z) / u
        
        if (l2% = 1) then d2t = (3.0 * t + 5.0 * x * dt + 4.0 * qz * qsqfm1) / u
        
        if (l3% = 1) then d3t = (8.0 * dt + 7.0 * x * d2t - 12.0 * qz * qz2 * x * qsqfm1) / u
        
      end if
      
    end if
    
  else
    
    ' compute by series
    
    u0i = 1.0
    
    if (l1% = 1) then u1i = 1.0
    
    if (l2% = 1) then u2i = 1.0
    
    if (l3% = 1) then u3i = 1.0
    
    term = 4.0
    
    tq = q * qsqfm1
    
    if (q < 0.5) then tqsum = 1.0 - q * qsq
    
    if (q >= 0.5) then tqsum = (1.0 / (1.0 + q) + q) * qsqfm1
    
    ttmold = term / 3.0
    
    t = ttmold * tqsum
    
    ' start of loop
    
    icounter% = 0
    
    do while(icounter% < n% or t <> told)
      
      icounter% = icounter% + 1
      
      p = icounter%
      
      u0i = u0i * u
      
      if (l1% = 1 and icounter% > 1) then u1i = u1i * u
      
      if (l2% = 1 and icounter% > 2) then u2i = u2i * u
      
      if (l3% = 1 and icounter% > 3) then u3i = u3i * u
      
      term = term * (p - 0.5) / p
      
      tq = tq * qsq
      
      tqsum = tqsum + tq
      
      told = t
      
      tterm = term / (2.0 * p + 3.0)
      
      tqterm = tterm * tqsum
      
      t = t - u0i * ((1.5 * p + 0.25) * tqterm / (p * p - 0.25) - ttmold * tq)
      
      ttmold = tterm
      
      tqterm = tqterm * p
      
      if (l1% = 1) then dt = dt + tqterm * u1i
      
      if (l2% = 1) then d2t = d2t + tqterm * u2i * (p - 1.0)
      
      if (l3% = 1) then d3t = d3t + tqterm * u3i * (p - 1.0) * (p - 2.0)
      
    loop
    
    if (l3% = 1) then d3t = 8.0 * x * (1.5 * d2t - xsq * d3t)
    
    if (l2% = 1) then d2t = 2.0 * (2.0 * xsq * d2t - dt)
    
    if (l1% = 1) then dt = -2.0 * x * dt
    
    t = t / xsq
    
  end if
  
end sub
  
  '''''''''''''''''''''''''''''''''''
  '''''''''''''''''''''''''''''''''''
  
sub xlamb(q, qsqfm1, tin, x, xpl, n%)
  
  ' Gooding lambert support subroutine
  
  ''''''''''''''''''''''''''''''''''''
  
  local tol = 3.0e-7, c0 = 1.7, xp1 = 0.0
  
  local c1 = 0.5, c2 = 0.03, c3 = 0.15, c41 = 1.0, c42 = 0.24
  
  thr2 = atan2(qsqfm1, 2.0 * q) / pi
  
  ' single-rev starter from t (at x = 0) & bilinear (usually)
  
  n% = 1
  
  tlamb(q, qsqfm1, 0.0, 0, dt, d2t, d3t, t0)
  
  tdiff = tin - t0
  
  if (tdiff <= 0.0) then
    
    x = t0 * tdiff / (-4.0 * tin)
    
    ' (-4 is the value of dt, for x = 0)
    
  else
    
    x = -tdiff / (tdiff + 4.0)
    
    w = x + c0 * sqr(2.0 * (1.0 - thr2))
    
    if (w < 0.0) then
      
      x = x - sqr(d8rt(-w)) * (x + sqr(tdiff / (tdiff + 1.5 * t0)))
      
    end if
    
    w = 4.0 / (4.0 + tdiff)
    
    x = x * (1.0 + x * (c1 * w - c2 * x * sqr(w)))
    
  end if
  
  ' (now have a starter, so proceed by halley)
  
  for i% = 1 to 3
    
    tlamb(q, qsqfm1, x, 2, dt, d2t, d3t, t)
    
    t = tin - t
    
    if (dt <> 0.0) then
      
      x = x + t * dt / (dt * dt + t * d2t / 2.0)
      
    end if
    
  next i%
  
end sub
  
  '''''''''''''''''''''''
  '''''''''''''''''''''''
  
function d8rt(x) as float
  
  ' Gooding lambert support function
  
  ''''''''''''''''''''''''''''''''''
  
  d8rt = sqr(sqr(sqr(x)))
  
end function
  
  ''''''''''''''''''''''''''''''''''''''
  ''''''''''''''''''''''''''''''''''''''
  
sub planet(ip%, xjdate, oev(), r(), v())
  
  ' heliocentric planetary ephemeris
  
  ' input
  
  '  ip%    = planet index (1 = mercury, 2 = venus, etc.)
  '  xjdate = julian date
  
  '  output
  
  '   r()   = heliocentric position vector (kilometers)
  '   v()   = heliocentric velocity vector (kilometers/second)
  '   oev() = classical orbital elements (kilometers & radians)
  
  '  note: coordinates are with respect to the
  '        ecliptic and equinox of J2000
  
  ''''''''''''''''''''''''''''''''''''''
  
  local t, t2, t3, argper, xma, eanom, tanom
  
  local sma, ecc, inc, mlong, raan, lper
  
  ' time arguments
  
  t = (xjdate - 2451545.0) / 36525.0
  
  t2 = t * t
  
  t3 = t * t2
  
  ' evaluate orbital elements for selected planet
  
  select case (ip%)
      
    case (1)
      
      ' mercury
      
      sma   =  0.387098310
      ecc   =  0.20563175 + 0.000020406 * t - 0.0000000284 * t2 - 0.00000000017 * t3
      inc   =  7.004986 - 0.0059516 * t + 0.00000081 * t2 + 0.000000041 * t3
      mlong =  252.250906 + 149472.6746358 * t - 0.00000535 * t2 + 0.000000002 * t3
      raan  =  48.330893 - 0.1254229 * t - 0.00008833 * t2 - 0.000000196 * t3
      lper  =  77.456119 + 0.1588643 * t - 0.00001343 * t2 + 0.000000039 * t3
      
    case (2)
      
      ' venus
      
      sma   =  0.723329820
      ecc   =  0.00677188 - 0.000047766 * t + 0.0000000975 * t2 + 0.00000000044 * t3
      inc   =  3.394622 - 0.0008568 * t - 0.00003244 * t2 + 0.000000010 * t3
      mlong =  181.979801 + 58517.815676 * t + 0.00000165 * t2 - 0.000000002 * t3
      raan  =  76.679920 - 0.2780080 * t - 0.00014256 * t2 - 0.000000198 * t3
      lper  =  131.563707 + 0.0048646 * t - 0.00138232 * t2 - 0.000005332 * t3
      
    case (3)
      
      ' earth
      
      sma   =  1.000001018
      ecc   =  0.01670862 - 0.000042037 * t - 0.0000001236 * t2 + 0.00000000004 * t3
      inc   =  0 + 0.0130546 * t - 0.00000931 * t2 - 0.000000034 * t3
      mlong =  100.466449 + 35999.3728519 * t - 0.00000568 * t2 - 0.000000000 * t3
      raan  =  174.873174 - 0.2410908 * t + 0.00004067 * t2 - 0.000001327 * t3
      lper  =  102.937348 + 0.3225557 * t + 0.00015026 * t2 + 0.000000478 * t3
      
    case (4)
      
      ' mars
      
      sma   =  1.523679342
      ecc   =  0.09340062 + 0.000090483 * t - 0.0000000806 * t2 - 0.00000000035 * t3
      inc   =  1.849726 - 0.0081479 * t - 0.00002255 * t2 - 0.000000027 * t3
      mlong =  355.433275 + 19140.2993313 * t + 0.00000261 * t2 - 0.000000003 * t3
      raan  =  49.558093 - 0.2949846 * t - 0.00063993 * t2 - 0.000002143 * t3
      lper  =  336.060234 + 0.4438898 * t - 0.00017321 * t2 + 0.000000300 * t3
      
    case (5)
      
      ' jupiter
      
      sma   =  5.202603191 + 0.0000001913 * t
      ecc   =  0.04849485 + 0.000163244 * t - 0.0000004719 * t2 - 0.00000000197 * t3
      inc   =  1.303270 - 0.0019872 * t + 0.00003318 * t2 + 0.000000092 * t3
      mlong =  34.351484 + 3034.9056746 * t - 0.00008501 * t2 + 0.000000004 * t3
      raan  =  100.464441 + 0.1766828 * t + 0.00090387 * t2 - 0.000007032 * t3
      lper  =  14.331309 + 0.2155525 * t + 0.00072252 * t2 - 0.000004590 * t3
      
    case (6)
      
      ' saturn
      
      sma   =  9.554909596 - 0.0000021389 * t
      ecc   =  0.05550862 - 0.000346818 * t - 0.0000006456 * t2 + 0.00000000338 * t3
      inc   =  2.488878 + 0.0025515 * t - 0.00004903 * t2 + 0.000000018 * t3
      mlong =  50.077471 + 1222.1137943 * t + 0.00021004 * t2 - 0.000000019 * t3
      raan  =  113.665524 - 0.2566649 * t - 0.00018345 * t2 + 0.000000357 * t3
      lper  =  93.056787 + 0.5665496 * t2 + 0.00052809 * t2 + 0.000004882 * t3
      
    case (7)
      
      ' uranus
      
      sma   =  19.218446062 - 0.0000000372 * t + 0.00000000098 * t2
      ecc   =  0.04629590 - 0.000027337 * t + 0.0000000790 * t2 + 0.00000000025 * t3
      inc   =  0.773196 - 0.0016869 * t + 0.00000349 * t2 + 0.000000016 * t3
      mlong =  314.055005 + 428.4669983 * t - 0.00000486 * t2 + 0.000000006 * t3
      raan  =  74.005947 + 0.0741461 * t + 0.00040540 * t2 + 0.000000104 * t3
      lper  =  173.005159 + 0.0893206 * t - 0.00009470 * t2 + 0.000000413 * t3
      
    case (8)
      
      ' neptune
      
      sma   =  30.110386869 - 0.0000001663 * t + 0.00000000069 * t2
      ecc   =  0.00898809 + 0.000006408 * t - 0.0000000008 * t2 - 0.00000000005 * t3
      inc   =  1.769952 + 0.0002257 * t + 0.00000023 * t2 - 0.000000000 * t3
      mlong =  304.348665 + 218.4862002 * t + 0.00000059 * t2 - 0.000000002 * t3
      raan  =  131.784057 - 0.0061651 * t - 0.00000219 * t2 - 0.000000078 * t3
      lper  =  48.123691 + 0.0291587 * t + 0.00007051 * t2 - 0.000000023 * t3
      
  end select
  
  ' argument of perihelion (radians)
  
  argper = modulo(dtr * (lper - raan))
  
  ' mean anomaly (radians)
  
  xma = modulo(dtr * (mlong - lper))
  
  ' solve Kepler's equation
  
  kepler1(xma, ecc, eanom, tanom)
  
  ' load orbital elements array
  
  oev(1) = aunit * sma
  oev(2) = ecc
  oev(3) = dtr * inc
  oev(4) = argper
  oev(5) = modulo(dtr * raan)
  oev(6) = tanom
  
  ' determine heliocentric state vector
  
  orb2eci(xmu, oev(), r(), v())
  
end sub
  
  ''''''''''''''''''''''''''''''''''''''''''''
  ''''''''''''''''''''''''''''''''''''''''''''
  
sub kepler1 (manom, ecc, eanom, tanom, niter%)
  
  ' solve Kepler's equation for circular,
  ' elliptic and hyperbolic orbits
  
  ' Danby's method
  
  ' input
  
  '  manom = mean anomaly (radians)
  '  ecc   = orbital eccentricity (non-dimensional)
  
  ' output
  
  '  eanom  = eccentric anomaly (radians)
  '  tanom  = true anomaly (radians)
  '  niter% = number of algorithm iterations
  
  ''''''''''''''''''''''''''''''''''''''''''
  
  local ktol, xma, s, c, f, fp, fpp, fppp
  
  local delta, deltastar, deltak, sta, cta
  
  ' define convergence criterion
  
  ktol = 1.0e-10
  
  xma = manom - pi2 * fix(manom/pi2)
  
  ' initial guess
  
  if (ecc = 0.0) then
    
    ' circular orbit
    
    tanom = xma
    
    eanom = xma
    
    return
    
  elseif (ecc < 1.0) then
    
    ' elliptic orbit
    
    eanom = xma + 0.85 * sgn(sin(xma)) * ecc
    
  else
    
    ' hyperbolic orbit
    
    eanom = log(2.0 * xma / ecc + 1.8)
    
  end if
  
  ' perform iterations
  
  niter% = 0
  
  do
    
    if (ecc < 1.0) then
      
      ' elliptic orbit
      
      s = ecc * sin(eanom)
      
      c = ecc * cos(eanom)
      
      f = eanom - s - xma
      
      fp = 1.0 - c
      
      fpp = s
      
      fppp = c
      
    else
      
      ' hyperbolic orbit
      
      s = ecc * sinh(eanom)
      
      c = ecc * cosh(eanom)
      
      f = s - eanom - xma
      
      fp = c - 1.0
      
      fpp = s
      
      fppp = c
      
    end if
    
    niter% = niter% + 1
    
    ' check for convergence
    
    if (abs(f) <= ktol or niter% > 20) then exit do
    
    ' update eccentric anomaly
    
    delta = -f / fp
    
    deltastar = -f / (fp + 0.5 * delta * fpp)
    
    deltak = -f / (fp + 0.5 * deltastar * fpp + deltastar * deltastar * fppp / 6.0)
    
    eanom = eanom + deltak
    
  loop
  
  if (niter% > 20) then
    
    print " "
    
    print "more than 20 iterations in kepler1"
    
    stop
    
  end if
  
  ' compute true anomaly
  
  if (ecc < 1.0) then
    
    ' elliptic orbit
    
    sta = sqr(1.0 - ecc * ecc) * sin(eanom)
    
    cta = cos(eanom) - ecc
    
  else
    
    ' hyperbolic orbit
    
    sta = sqr(ecc * ecc - 1.0) * sinh(eanom)
    
    cta = ecc - cosh(eanom)
    
  end if
  
  tanom = atan3(sta, cta)
  
end sub
  
  ''''''''''''''''''''''''''''''
  ''''''''''''''''''''''''''''''
  
sub eci2orb(mu, r(), v(), oev())
  
  ' convert eci state vector to classical orbital elements
  
  ' input
  
  '  mu  = central body gravity (km^3/sec^2)
  '  r() = eci position vector (kilometers)
  '  v() = eci velocity vector (kilometers/second)
  
  ' output
  
  '  oev(1)  = semimajor axis (kilometers)
  '  oev(2)  = orbital eccentricity (non-dimensional)
  '            (0 <= eccentricity < 1)
  '  oev(3)  = orbital inclination (radians)
  '            (0 <= inclination <= pi)
  '  oev(4)  = argument of perigee (radians)
  '            (0 <= argument of perigee <= 2 pi)
  '  oev(5)  = right ascension of ascending node (radians)
  '            (0 <= raan <= 2 pi)
  '  oev(6)  = true anomaly (radians)
  '            (0 <= true anomaly <= 2 pi)
  
  ''''''''''''''''''''''''''''''''''''''''
  
  local rmag, vmag, sma, p, q, const1
  
  local h, xk, x1, y1, eccm, inc, xlambdat
  
  local raan, argper, tanom
  
  local rhat(3), vhat(3), hhat(3), vtmp(3), ecc(3)
  
  local hv(3), fhat(3), ghat(3)
  
  ' position and velocity magnitude
  
  rmag = vecmag(r())
  
  vmag = vecmag(v())
  
  ' position and velocity unit vectors
  
  uvector(r(), rhat())
  
  uvector(v(), vhat())
  
  ' angular momentum vectors
  
  vcross(r(), v(), hv())
  
  uvector(hv(), hhat())
  
  ' eccentricity vector
  
  for i% = 1 to 3
    
    vtmp(i%) = v(i%) / mu
    
  next i%
  
  vcross(vtmp(), hv(), ecc())
  
  for i% = 1 to 3
    
    ecc(i%) = ecc(i%) - rhat(i%)
    
  next i%
  
  ' semimajor axis
  
  sma = 1.0 / (2.0 / rmag - vmag^2 / mu)
  
  p = hhat(1) / (1.0 + hhat(3))
  
  q = -hhat(2) / (1.0 + hhat(3))
  
  const1 = 1.0 / (1.0 + p^2 + q^2)
  
  fhat(1) = const1 * (1.0 - p^2 + q^2)
  fhat(2) = const1 * 2.0 * p * q
  fhat(3) = -const1 * 2.0 * p
  
  ghat(1) = const1 * 2.0 * p * q
  ghat(2) = const1 * (1.0 + p^2 - q^2)
  ghat(3) = const1 * 2.0 * q
  
  h = vdot(ecc(), ghat())
  
  xk = vdot(ecc(), fhat())
  
  x1 = vdot(r(), fhat())
  
  y1 = vdot(r(), ghat())
  
  ' orbital eccentricity
  
  eccm = sqr(h * h + xk * xk)
  
  ' orbital inclination
  
  inc = 2.0 * atn(sqr(p * p + q * q))
  
  ' true longitude
  
  xlambdat = atan3(y1, x1)
  
  ' check for equatorial orbit
  
  if (inc > 1.0e-10) then
    
    raan = atan3(p, q)
    
  else
    
    raan = 0.0
    
  end if
  
  ' check for circular orbit
  
  if (eccm > 1.0e-10) then
    
    argper = modulo(atan3(h, xk) - raan)
    
  else
    
    argper = 0.0
    
  end if
  
  ' true anomaly
  
  tanom = modulo(xlambdat - raan - argper)
  
  oev(1) = sma
  oev(2) = eccm
  oev(3) = inc
  oev(4) = argper
  oev(5) = raan
  oev(6) = tanom
  
end sub
  
  ''''''''''''''''''''''''''''''
  ''''''''''''''''''''''''''''''
  
sub orb2eci(mu, oev(), r(), v())
  
  ' convert orbital elements to state vector subroutine
  
  ' input
  
  '  mu     = gravitational constant (km^3/sec^2)
  '  oev(1) = semimajor axis (kilometers)
  '  oev(2) = orbital eccentricity (non-dimensional)
  '  oev(3) = orbital inclination (radians)
  '           (0 <= oev(3) <= pi)
  '  oev(4) = argument of perigee (radians)
  '           (0 <= oev(4) <= 2 pi)
  '  oev(5) = right ascension of ascending node or
  '           east longitude of ascending node (radians)
  '           (0 <= oev(5) <= 2 pi)
  '  oev(6) = true anomaly (radians)
  '           (0 <= oev(6) <= 2 pi)
  
  ' output
  
  '  r() = geocentric, equatorial position vector (kilometers)
  '  v() = geocentric, equatorial velocity vector (kilometers/second)
  
  ' special notes
  
  '  (1) eci coordinates if oev(5) = raan
  
  '  (2) ecf coordinates if oev(5) = east longitude
  
  '  (3) valid for elliptic and hyperbolic orbits
  
  '''''''''''''''''''''''''''''''''''''''''''''''
  
  local sma, ecc, inc, argper, nangle, tanom
  
  local slr, rm, arglat, sarglat, carglat
  
  local sinc, cinc, snangle, cnangle, c1, c2, c3
  
  ' "unload" orbital elements array
  
  sma = oev(1)
  ecc = oev(2)
  inc = oev(3)
  argper = oev(4)
  nangle = oev(5)
  tanom = oev(6)
  
  ' semiparameter
  
  slr = sma * (1.0 - ecc * ecc)
  
  ' radius magnitude
  
  rm = slr / (1.0 + ecc * cos(tanom))
  
  ' argument of latitude
  
  arglat = argper + tanom
  
  sarglat = sin(arglat)
  
  carglat = cos(arglat)
  
  sinc = sin(inc)
  
  cinc = cos(inc)
  
  snangle = sin(nangle)
  
  cnangle = cos(nangle)
  
  c1 = sqr(mu / slr)
  
  c2 = ecc * cos(argper) + carglat
  
  c3 = ecc * sin(argper) + sarglat
  
  ' x, y and z components of the position vector
  
  r(1) = rm * (cnangle * carglat - snangle * cinc * sarglat)
  
  r(2) = rm * (snangle * carglat + cinc * sarglat * cnangle)
  
  r(3) = rm * sinc * sarglat
  
  ' x, y and z components of the velocity vector
  
  v(1) = -c1 * (cnangle * c3 + snangle * cinc * c2)
  
  v(2) = -c1 * (snangle * c3 - cnangle * cinc * c2)
  
  v(3) = c1 * c2 * sinc
  
end sub
  
  ''''''''''''''''''''''''''''''''
  ''''''''''''''''''''''''''''''''
  
sub julian(month, day, year, jday)
  
  ' Gregorian date to julian day subroutine
  
  ' input
  
  '  month = calendar month
  '  day   = calendar day
  '  year  = calendar year (all four digits)
  
  ' output
  
  '  jday = julian day
  
  ' special notes
  
  '  (1) calendar year must include all digits
  
  '  (2) will report October 5, 1582 to October 14, 1582
  '      as invalid calendar dates and exit
  
  '''''''''''''''''''''''''''''''''''''''''
  
  local a, b, c, m, y
  
  y = year
  
  m = month
  
  b = 0.0
  
  c = 0.0
  
  if (m <= 2.0) then
    
    y = y - 1.0
    
    m = m + 12.0
    
  end if
  
  if (y < 0.0) then c = -0.75
  
  if (year < 1582.0) then
    
    ' null
    
  elseif (year > 1582.0) then
    
    a = fix(y / 100.0)
    
    b = 2.0 - a + fix(a / 4.0)
    
  elseif (month < 10.0) then
    
    ' null
    
  elseif (month > 10.0) then
    
    a = fix(y / 100.0)
    
    b = 2.0 - a + fix(a / 4.0)
    
  elseif (day <= 4.0) then
    
    ' null
    
  elseif (day > 14.0) then
    
    a = fix(y / 100.0)
    
    b = 2.0 - a + fix(a / 4.0)
    
  else
    
    print "this date does not exist!!"
    
    exit
    
  end if
  
  jday = fix(365.25 * y + c) + fix(30.6001 * (m + 1.0)) + day + b + 1720994.5
  
end sub
  
  ''''''''''''''''''''''''''''''''
  ''''''''''''''''''''''''''''''''
  
sub gdate (jday, month, day, year)
  
  ' Julian day to Gregorian date subroutine
  
  ' input
  
  '  jday = julian day
  
  ' output
  
  '  month = calendar month
  '  day   = calendar day
  '  year  = calendar year
  
  ''''''''''''''''''''''''
  
  local a, b, c, d, e, f, z, alpha
  
  z = fix(jday + 0.5)
  
  f = jday + 0.5 - z
  
  if (z < 2299161) then
    
    a = z
    
  else
    
    alpha = fix((z - 1867216.25) / 36524.25)
    
    a = z + 1.0 + alpha - fix(alpha / 4.0)
    
  end if
  
  b = a + 1524.0
  
  c = fix((b - 122.1) / 365.25)
  
  d = fix(365.25 * c)
  
  e = fix((b - d) / 30.6001)
  
  day = b - d - fix(30.6001 * e) + f
  
  if (e < 13.5) then
    
    month = e - 1.0
    
  else
    
    month = e - 13.0
    
  end if
  
  if (month > 2.5) then
    
    year = c - 4716.0
    
  else
    
    year = c - 4715.0
    
  end if
  
end sub
  
  ''''''''''''''''''''''''
  ''''''''''''''''''''''''
  
sub utc2tdb (jdutc, jdtdb)
  
  ' convert UTC julian date to TDB julian date
  
  ' input
  
  '  jdutc   = UTC julian day
  
  ' output
  
  '  jdtdb = TDB julian day
  
  ' Reference Frames in Astronomy and Geophysics
  ' J. Kovalevsky et al., 1989, pp. 439-442
  
  '''''''''''''''''''''''''''''''''''''''''
  
  local corr, jdtt, t, leapsecond
  
  ' find current number of leap seconds
  
  findleap(jdutc, leapsecond)
  
  ' compute TDT julian date
  
  corr = (leapsecond + 32.184) / 86400.0
  
  jdtt = jdutc + corr
  
  ' time argument for correction
  
  t = (jdtt - 2451545.0) / 36525.0
  
  ' compute correction in microseconds
  
  corr = 1656.675 * sin(dtr * (35999.3729 * t + 357.5287))
  corr = corr + 22.418     * sin(dtr * (32964.467  * t + 246.199))
  corr = corr + 13.84      * sin(dtr * (71998.746  * t + 355.057))
  corr = corr +  4.77      * sin(dtr * ( 3034.906  * t +  25.463))
  corr = corr +  4.677     * sin(dtr * (34777.259  * t + 230.394))
  corr = corr + 10.216 * t * sin(dtr * (35999.373  * t + 243.451))
  corr = corr +  0.171 * t * sin(dtr * (71998.746  * t + 240.98 ))
  corr = corr +  0.027 * t * sin(dtr * ( 1222.114  * t + 194.661))
  corr = corr +  0.027 * t * sin(dtr * ( 3034.906  * t + 336.061))
  corr = corr +  0.026 * t * sin(dtr * (  -20.186  * t +   9.382))
  corr = corr +  0.007 * t * sin(dtr * (29929.562  * t + 264.911))
  corr = corr +  0.006 * t * sin(dtr * (  150.678  * t +  59.775))
  corr = corr +  0.005 * t * sin(dtr * ( 9037.513  * t + 256.025))
  corr = corr +  0.043 * t * sin(dtr * (35999.373  * t + 151.121))
  
  ' convert corrections to days
  
  corr = 0.000001 * corr / 86400.0
  
  ' TDB julian date
  
  jdtdb = jdtt + corr
  
end sub
  
  '''''''''''''''
  '''''''''''''''
  
sub jd2str(jdutc)
  
  ' convert julian day to calendar date and UTC time
  
  ''''''''''''''''''''''''''''''''''''''''''''''''''
  
  gdate (jdutc, cmonth, day, year)
  
  print "calendar date  ", month$(cmonth) + " " + STR$(int(day)) + " " + str$(year)
  
  print " "
  
  thr0 = 24.0 * (day - int(day))
  
  thr = int(thr0)
  
  tmin0 = 60.0 * (thr0 - thr)
  
  tmin = int(tmin0)
  
  tsec = 60.0 * (tmin0 - tmin)
  
  ' fix seconds and minutes for rollover
  
  if (tsec >= 60.0) then
    
    tsec = 0.0
    
    tmin = tmin + 1.0
    
  end if
  
  ' fix minutes for rollover
  
  if (tmin >= 60.0) then
    
    tmin = 0.0
    
    thr = thr + 1.0
    
  end if
  
  print "UTC time       ", str$(thr) + " hours " + str$(tmin) + " minutes " + str$(tsec, 0, 2) + " seconds"
  
end sub
  
  ''''''''''''''''''''''''''''
  ''''''''''''''''''''''''''''
  
sub findleap(jday, leapsecond)
  
  ' find number of leap seconds for utc julian day
  
  ' input
  
  '  jday = utc julian day
  
  ' input via global
  
  '  jdleap  = array of utc julian dates
  '  leapsec = array of leap seconds
  
  ' output
  
  '  leapsecond = number of leap seconds
  
  ''''''''''''''''''''''''''''''''''''''
  
  if (jday <= jdleap(1)) then
    
    ' date is <= 1972; set to first data element
    
    leapsecond = leapsec(1)
    
    exit sub
    
  end if
  
  if (jday >= jdleap(28)) then
    
    ' date is >= end of current data
    ' set to last data element
    
    leapsecond = leapsec(28)
    
    exit sub
    
  end if
  
  ' find data within table
  
  for i% = 1 to 27
    
    if (jday >= jdleap(i%) and jday < jdleap(i% + 1)) then
      
      leapsecond = leapsec(i%)
      
      exit sub
      
    end if
    
  next i%
  
end sub
  
  ''''''''''''''''''''''''''''
  ''''''''''''''''''''''''''''
  
sub getdate (month, day, year)
  
  ' request calendar date subroutine
  
  do
    print " "
    print "please input the calendar date"
    print " "
    print "(month [1 - 12], day [1 - 31], year [yyyy])"
    print "< for example, october 21, 1986 is input as 10,21,1986 >"
    print "< b.c. dates are negative, a.d. dates are positive >"
    print "< the day of the month may also include a decimal part >"
    print " "
    input month, day, year
    
  loop until (month >= 1 and month <= 12) and (day >= 1 and day <= 31)
  
end sub
  
  ''''''''''''''''''''''''''
  ''''''''''''''''''''''''''
  
sub getutc (thr, tmin, tsec)
  
  ' request time subroutine
  
  do
    print " "
    print "please input the UTC time"
    print " "
    print "(hours [0 - 24], minutes [0 - 60], seconds [0 - 60])"
    print " "
    input thr, tmin, tsec
    
  loop until ((thr >= 0.0 and thr <= 24.0) and (tmin >= 0.0 and tmin <= 60.0) and (tsec >= 0.0 and tsec <= 60.0))
  
end sub
  
  ''''''''''''''''
  ''''''''''''''''
  
sub oeprint(oev())
  
  ' print classical orbital elements subroutine
  
  '''''''''''''''''''''''''''''''''''''''''''''
  
  print "semimajor axis          ", str$((oev(1) / aunit), 0, 10) + " AU"
  print "eccentricity            ", str$(oev(2), 0, 10)
  print "orbital inclination     ", str$(rtd * oev(3), 0, 4) + " degrees"
  print "argument of perigee     ", str$(rtd * oev(4), 0, 4) + " degrees"
  print "RAAN                    ", str$(rtd * oev(5), 0, 4) + " degrees"
  print "true anomaly            ", str$(rtd * oev(6), 0, 4) + " degrees"
  print " "
  
end sub
  
  '''''''''''''''''''
  '''''''''''''''''''
  
sub svprint(r(), v())
  
  ' print position and velocity vectors subroutine
  
  ''''''''''''''''''''''''''''''''''''''''''''''''
  
  local rmag, vmag
  
  print "position vector x-component  ", str$(r(1), 0, -12), " kilometers"
  print "position vector y-component  ", str$(r(2), 0, -12), " kilometers"
  print "position vector z-component  ", str$(r(3), 0, -12), " kilometers"
  
  rmag = vecmag(r())
  
  print " "
  
  print "position vector magnitude    ", str$(rmag, 0, -12), " kilometers"
  
  print " "
  
  print "velocity vector x-component  ", str$(v(1), 0, -12), " kilometers/second"
  print "velocity vector y-component  ", str$(v(2), 0, -12), " kilometers/second"
  print "velocity vector z-component  ", str$(v(3), 0, -12), " kilometers/second"
  
  vmag = vecmag(v())
  
  print " "
  
  print "velocity vector magnitude    ", str$(vmag, 0, -12), " kilometers/second"
  
end sub
  
  '''''''''''''''''''''''
  '''''''''''''''''''''''
  
sub usr_func (n%, x(), f)
  
  ' objective function subroutine
  
  '''''''''''''''''''''''''''''''
  
  local i%, jdate1, jdate2, tof
  
  local ri(3), vi(3), rf(3), vf(3)
  
  local sv1(6), sv2(6)
  
  local dvm1, dvm2
  
  jdate1 = x(1) + jdate0
  
  jdate2 = x(2) + jdate0
  
  ' compute initial state vector
  
  planet(ip1%, jdate1, oevi(), ri(), vi())
  
  ' compute final state vector
  
  planet(ip2%, jdate2, oevf(), rf(), vf())
  
  'svprint(rf(), vf())
  
  tof = (jdate2 - jdate1) * 86400.0
  
  ' set up for Gooding's method
  
  for i% = 1 to 3
    
    sv1(i%) = ri(i%)
    
    sv2(i%) = rf(i%)
    
    sv1(i% + 3) = vi(i%)
    
    sv2(i% + 3) = vf(i%)
    
  next i%
  
  ' solve Lambert's TPBVP using Gooding's method
  
  glambert(xmu, sv1(), sv2(), tof, vito(), vfto())
  
  ' calculate departure delta-v (kilometers/second)
  
  dv1(1) = vito(1) - vi(1)
  dv1(2) = vito(2) - vi(2)
  dv1(3) = vito(3) - vi(3)
  
  dvm1 = vecmag(dv1())
  
  ' calculate arrival delta-v (kilometers/second)
  
  dv2(1) = vf(1) - vfto(1)
  dv2(2) = vf(2) - vfto(2)
  dv2(3) = vf(3) - vfto(3)
  
  dvm2 = vecmag(dv2())
  
  ' load scalar objective function
  
  select case otype%
      
    case 1
      
      ' launch delta-v
      
      f = dvm1
      
    case 2
      
      ' arrival delta-v
      
      f = dvm2
      
    case 3
      
      ' launch + arrival delta-v
      
      f = dvm1 + dvm2
      
  end select
  
end sub
  
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  
sub qnopt (n%, tol, maxiter%, x(), f, niter%, ier%, nsig%)
  
  ' quasi-newton multi-dimensional minimization subroutine
  
  ' input
  '  n%       = number of variables
  '  tol      = convergence criteria
  '  maxiter% = maximum number of function evaluations
  '  x()      = current guess for solution
  '             (n% rows by 1 column)
  
  ' output
  '  x()    = solution vector (n% rows by 1 column)
  '  f      = current objective function value
  '  niter% = number of function evaluations
  '  ier%   = error indicator
  '  nsig%  = number of significant digits
  
  ' note: requires function log10 and subroutines
  '       utility and usr_func
  
  ''''''''''''''''''''
  
  local h(n% * (n% + 1) / 2), w(3 * n%), g(n%)
  
  local i%, ax, eps, p1
  
  ' initialization
  
  ax = 0.1
  
  eps = 2.23e-16
  
  p1 = 0.1
  
  ier% = 0
  
  hh = sqr(eps)
  
  ig% = n%
  igg% = n% + n%
  iss% = igg%
  idiff% = 1
  irr% = n%
  
  w(1) = -1.0
  w(2) = 0.0
  w(3) = 0.0
  
  for i% = 1 to n%
    
    g(i%) = x(i%)
    
  next i%
  
  usr_func(n%, g(), f)
  
  ifn% = 1
  
  if (n% = 1) then
    
    ' null
    
  else
    
    ij% = 2
    
    for i% = 2 to n%
      
      for j% = 2 to i%
        
        h(ij%) = 0.0
        
        ij% = ij% + 1
        
      next j%
      
      ij% = ij% + 1
      
    next i%
    
  end if
  
  ij% = 0
  
  for i% = 1 to n%
    
    ij% = ij% + i%
    
    h(ij%) = 1.0
    
  next i%
  
  itn% = 0
  
  df = -1.0
  
qn1:
  
  link% = 1
  
  goto qn18
  
qn2:
  
  if (ifn% >= maxiter%) then goto qn14
  
  itn% = itn% + 1
  
  for i% = 1 to n%
    
    w(i%) = -w(ig% + i%)
    
  next i%
  
  if (irr% < n%) then goto qn4
  
  g(1) = w(1)
  
  if (n% > 1) then goto qn3
  
  w(1) = w(1) / h(1)
  
  goto qn4
  
qn3:
  
  ii% = 1
  
  for i% = 2 to n%
    
    ij% = ii%
    
    ii% = ii% + i%
    
    v = w(i%)
    
    im1% = i% - 1
    
    for j% = 1 to im1%
      
      ij% = ij% + 1
      
      v = v - h(ij%) * w(j%)
      
    next j%
    
    g(i%) = v
    
    w(i%) = v
    
  next i%
  
  w(n%) = w(n%) / h(ii%)
  
  jj% = ii%
  
  nm1% = n% - 1
  
  for nj% = 1 to nm1%
    
    j% = n% - nj%
    
    jp1% = j% + 1
    
    jj% = jj% - jp1%
    
    v = w(j%) / h(jj%)
    
    ij% = jj%
    
    for i% = jp1% to n%
      
      ij% = ij% + i% - 1
      
      v = v - h(ij%) * w(i%)
      
    next i%
    
    w(j%) = v
    
  next nj%
  
qn4:
  
  relx = 0.0
  
  gs0 = 0.0
  
  for i% = 1 to n%
    
    w(iss% + i%) = w(i%)
    
    tmp1 = abs(x(i%))
    
    tmp2 = ax
    
    if (tmp1 > tmp2) then
      
      tmp = tmp1
      
    else
      
      tmp = tmp2
      
    end if
    
    diff = abs(w(i%)) / tmp
    
    tmp1 = relx
    
    tmp2 = diff
    
    if (tmp1 > tmp2) then
      
      tmp = tmp1
      
    else
      
      tmp = tmp2
      
    end if
    
    relx = tmp
    
    gs0 = gs0 + w(ig% + i%) * w(i%)
    
  next i%
  
  if (relx = 0.0) then goto qn15
  
  aeps = tol / relx
  
  ier% = 2
  
  if (gs0 >= 0.0 or df = 0.0) then goto qn15
  
  ier% = 0
  
  alpha = (-df - df) / gs0
  
  if (alpha <= 0.0) then alpha = 1.0
  
  tmp1 = alpha
  
  tmp2 = 1.0
  
  if (tmp1 < tmp2) then
    
    tmp = tmp1
    
  else
    
    tmp = tmp2
    
  end if
  
  alpha = tmp
  
  if (idiff% = 2) then
    
    tmp1 = p1
    
    tmp2 = alpha
    
    if (tmp1 > tmp2) then
      
      tmp = tmp1
      
    else
      
      tmp = tmp2
      
    end if
    
    alpha = tmp
    
  end if
  
  ff = f
  tot = 0.0
  jnt% = 0
  
qn5:
  
  if (ifn% >= maxiter%) then goto qn14
  
  for i% = 1 to n%
    
    w(i%) = x(i%) + alpha * w(iss% + i%)
    
  next i%
  
  usr_func(n%, w(), f1)
  
  ifn% = ifn% + 1
  
  if (f1 >= f) then goto qn8
  
  f2 = f
  
  tot = tot + alpha
  
qn6:
  
  ier% = 0
  f = f1
  
  for i% = 1 to n%
    
    x(i%) = w(i%)
    
  next i%
  
  if (jnt% - 1) < 0 then goto qn7
  
  if (jnt% - 1) = 0 then goto qn10
  
  if (jnt% - 1) > 0 then goto qn11
  
qn7:
  
  if (ifn% >= maxiter%) then goto qn14
  
  for i% = 1 to n%
    
    w(i%) = x(i%) + alpha * w(iss% + i%)
    
  next i%
  
  usr_func(n%, w(), f1)
  
  ifn% = ifn% + 1
  
  if (f1 >= f) then goto qn11
  
  if (f1 + f2 >= f + f and 7.0 * f1 + 5.0 * f2 > 12.0 * f) then jnt% = 2
  
  tot = tot + alpha
  
  alpha = alpha + alpha
  
  goto qn6
  
qn8:
  
  if (f = ff and idiff% = 2 and relx > tol) then ier% = 2
  
  if (alpha < aeps) then goto qn15
  
  if (ifn% >= maxiter%) then goto qn14
  
  alpha = 0.5 * alpha
  
  for i% = 1 to n%
    
    w(i%) = x(i%) + alpha * w(iss% + i%)
    
  next i%
  
  usr_func(n%, w(), f2)
  
  ifn% = ifn% + 1
  
  if (f2 >= f) then goto qn9
  
  tot = tot + alpha
  
  ier% = 0
  
  f = f2
  
  for i% = 1 to n%
    
    x(i%) = w(i%)
    
  next i%
  
  goto qn10
  
qn9:
  
  z = p1
  
  if (f1 + f > f2 + f2) then z = 1.0 + 0.5 * (f - f1) / (f + f1 - f2 - f2)
  
  tmp1 = p1
  
  tmp2 = z
  
  if (tmp1 > tmp2) then
    
    tmp = tmp1
    
  else
    
    tmp = tmp2
    
  end if
  
  z = tmp
  
  alpha = z * alpha
  
  jnt% = 1
  
  goto qn5
  
qn10:
  
  if (tot < aeps) then goto qn15
  
qn11:
  
  alpha = tot
  
  for i% = 1 to n%
    
    w(i%) = w(ig% + i%)
    
  next i%
  
  link% = 2
  
  goto qn18
  
qn12:
  
  if (ifn% >= maxiter%) then goto qn14
  
  gys = 0.0
  
  for i% = 1 to n%
    
    gys = gys + w(ig% + i%) * w(iss% + i%)
    
    w(igg% + i%) = w(i%)
    
  next i%
  
  df = ff - f
  
  dgs = gys - gs0
  
  if (dgs <= 0.0) then goto qn2
  
  if (dgs + alpha * gs0 > 0.0) then goto qn13
  
  sig = 1.0 / gs0
  
  irr% = -irr%
  
  utility(h(), n%, w(), sig, g(), irr%, 0, 0.0)
  
  for i% = 1 to n%
    
    g(i%) = w(ig% + i%) - w(igg% + i%)
    
  next i%
  
  sig = 1.0 / (alpha * dgs)
  
  irr% = -irr%
  
  utility(h(), n%, g(), sig, w(), irr%, 0, 0.0)
  
  goto qn2
  
qn13:
  
  zz = alpha / (dgs - alpha * gs0)
  sig = -zz
  
  utility(h(), n%, w(), sig, g(), irr%, 0, eps)
  
  z = dgs * zz - 1.0
  
  for i% = 1 to n%
    
    g(i%) = w(ig% + i%) + z * w(igg% + i%)
    
  next i%
  
  sig = 1.0 / (zz * dgs * dgs)
  
  utility(h(), n%, g(), sig, w(), irr%, 0, 0.0)
  
  goto qn2
  
qn14:
  
  ier% = 3
  
  goto qn16
  
qn15:
  
  if (idiff% = 2) then goto qn16
  
  idiff% = 2
  
  goto qn1
  
qn16:
  
  if (ier% <> 0) then goto qn17
  
  if (relx <= tol) then goto qn17
  
  goto qn1
  
qn17:
  
  gnrm = 0.0
  
  for i% = 1 to n%
    
    g(i%) = w(ig% + i%)
    
    gnrm = gnrm + g(i%) * g(i%)
    
  next i%
  
  gnrm = sqr(gnrm)
  w(1) = gnrm
  w(2) = ifn%
  
  tmp1 = eps
  
  tmp2 = relx
  
  if (tmp1 > tmp2) then
    
    tmp = tmp1
    
  else
    
    tmp = tmp2
    
  end if
  
  w(3) = -log10(tmp)
  
  if (n% = 1) then goto exitsub
  
  np1% = n% + 1
  
  nm1% = n% - 1
  
  jj% = (n% * (np1%)) / 2
  
  for jb% = 1 to nm1%
    
    jp1% = np1% - jb%
    
    jj% = jj% - jp1%
    
    hjj = h(jj%)
    
    ij% = jj%
    
    l% = 0
    
    for i% = jp1% to n%
      l% = l% + 1
      ij% = ij% + i% - 1
      v = h(ij%) * hjj
      kj% = ij%
      for k% = i% to n%
        h(kj% + l%) = h(kj% + l%) + h(kj%) * v
        kj% = kj% + k%
      next k%
      h(ij%) = v
    next i%
    hjj = h(jj%)
  next jb%
  
  goto exitsub
  
qn18:
  
  if (idiff% = 2) then goto qn19
  
  for i% = 1 to n%
    tmp1 = abs(x(i%))
    tmp2 = ax
    if (tmp1 > tmp2) then
      tmp = tmp1
    else
      tmp = tmp2
    end if
    z = hh * tmp
    zz = x(i%)
    x(i%) = zz + z
    usr_func(n%, x(), f1)
    w(ig% + i%) = (f1 - f) / z
    x(i%) = zz
  next i%
  
  ifn% = ifn% + n%
  
  if (link% = 1) then
    goto qn2
  elseif (link% = 2) then
    goto qn12
  end if
  
qn19:
  
  for i% = 1 to n%
    tmp1 = abs(x(i%))
    tmp2 = ax
    if (tmp1 > tmp2) then
      tmp = tmp1
    else
      tmp = tmp2
    end if
    z = hh * tmp
    zz = x(i%)
    x(i%) = zz + z
    usr_func(n%, x(), f1)
    x(i%) = zz - z
    usr_func(n%, x(), f2)
    w(ig% + i%) = (f1 - f2) / (z + z)
    x(i%) = zz
  next i%
  
  ifn% = ifn% + n% + n%
  
  if (link% = 1) then
    goto qn2
  elseif (link% = 2) then
    goto qn12
  end if
  
exitsub:
  
  ' total number of function calls
  
  niter% = ifn%
  
  ' number of significant digits
  
  nsig% = w(3)
  
end sub
  
  '''''''''''''''''''''''''''''''''''''''''''''''''
  
sub utility (a(), n%, z(), sig, w(), irr%, mk%, tol)
  
  ' optimization utility subroutine
  
  '''''''''''''''''''''''''''''''''
  
  if (n% > 1) then
    
    ' null
    
  else
    
    a(1) = a(1) + sig * z(1) * z(1)
    
    ir% = 1
    
    if (a(1) > 0.0) then exit sub
    
    a(1) = 0.0
    
    irr% = 0
    
    exit sub
    
  end if
  
  if (sig > 0.0) then goto util1
  
  if (sig = 0.0 or ir% = 0) then exit sub
  
  ti = 1.0 / sig
  
  jj% = 0
  
  if (mk% = 0) then
    
    for j% = 1 to n%
      
      w(j%) = z(j%)
      
    next j%
    
    for j% = 1 to n%
      
      jj% = jj% + j%
      
      v = w(j%)
      
      if (a(jj%) > 0.0) then
        
        ti = ti + (v * v) / a(jj%)
        
        if (j% <> n%) then
          
          ij% = jj%
          
          jp1% = j% + 1
          
          for i% = jp1% to n%
            
            ij% = ij% + i% - 1
            
            w(i%) = w(i%) - v * a(ij%)
            
          next i%
          
        end if
        
      else
        
        w(j%) = 0.0
        
      end if
      
    next j%
    
  else
    
    for j% = 1 to n%
      
      jj% = jj% + j%
      
      if (a(jj%) <> 0.0) then ti = ti + (w(j%) * w(j%)) / a(jj%)
      
    next j%
    
  end if
  
  if (irr% <= 0) then
    
    ti = 0.0
    
    irr% = -irr% - 1
    
  elseif (ti > 0.0) then
    
    ti = tol / sig
    
    if (tol = 0.0) then irr% = irr% - 1
    
  elseif (mk% - 1) <= 0 then
    
    goto util1
    
  end if
  
  tim = ti
  ii% = jj%
  i% = n%
  
  for j% = 1 to n%
    
    if (a(ii%) <> 0.0) then tim = ti - (w(i%) * w(i%)) / a(ii%)
    
    w(i%) = ti
    
    ti = tim
    
    ii% = ii% - i%
    
    i% = i% - 1
    
  next j%
  
  mm% = 1
  
  goto util2
  
util1:
  
  mm% = 0
  
  tim = 1.0 / sig
  
util2:
  
  jj% = 0
  
  for j% = 1 to n%
    
    jj% = jj% + j%
    
    ij% = jj%
    
    jp1% = j% + 1
    
    v = z(j%)
    
    if (a(jj%) = 0.0) then
      
      if (ir% > 0 or sig < 0.0 or v = 0.0) then
        
        ' null
        
      else
        
        irr% = 1 - irr%
        
        a(jj%) = (v * v) / tim
        
        if (j% = n%) then exit sub
        
        for i% = jp1% to n%
          
          ij% = ij% + i% - 1
          
          a(ij%) = z(i%) / v
          
        next i%
        
        exit sub
        
      end if
      
      ti = tim
      
    else
      
      al = v / a(jj%)
      
      ti = w(j%)
      
      if (mm% = 0) then ti = tim + v * al
      
      r = ti / tim
      
      a(jj%) = r * a(jj%)
      
      if (r = 0.0 or j% = n%) then exit for
      
      b = al / ti
      
      if (r > 4.0) then
        
        gm = tim / ti
        
        for i% = jp1% to n%
          
          ij% = ij% + i% - 1
          
          y = a(ij%)
          
          a(ij%) = b * z(i%) + y * gm
          
          z(i%) = z(i%) - v * y
          
        next i%
        
      else
        
        for i% = jp1% to n%
          
          ij% = ij% + i% - 1
          
          z(i%) = z(i%) - v * a(ij%)
          
          a(ij%) = a(ij%) + b * z(i%)
          
        next i%
        
      end if
      
      tim = ti
      
    end if
    
  next j%
  
  if (irr% < 0) then irr% = -irr%
  
end sub
  
  '''''''''''''''''''''''''
  '''''''''''''''''''''''''
  
function modulo(x) as float
  
  ' modulo 2 pi function
  
  ''''''''''''''''''''''
  
  local a
  
  a = x - pi2 * fix(x / pi2)
  
  if (a < 0.0) then
    
    a = a + pi2
    
  end if
  
  modulo = a
  
end function

  '''''''''''''''''''''''''''
  '''''''''''''''''''''''''''

function atan2(y, x) as float

  ' four quadrant inverse tangent function
  
  ' input
  
  '  x = sine of angle
  '  y = cosine of angle
  
  ' output
  
  '  atan2 = angle (-pi =< atan2 <= pi; radians)
  
  ''''''''''''''''''''''''''''''''''''''''''''''''
  
  local ax, ay, phi
    
  if (x = 0.0) and (y = 0.0) then
    
    atan2 = 0.0
    
  else
    
    ax = abs(x)
    
    ay = abs(y)
    
    if (ax > ay) then
      
      phi = atn(ay / ax)
      
    else
      
      phi = pidiv2 - atn(ax / ay)
      
    end if
    
    if (x < 0.0) then phi = pi - phi
    
    if (y < 0.0) then phi = - phi
    
    atan2 = phi
    
  end if
  
end function
  
  '''''''''''''''''''''''''''
  '''''''''''''''''''''''''''
  
function atan3(a, b) as float
  
  ' four quadrant inverse tangent function
  
  ' input
  
  '  a = sine of angle
  '  b = cosine of angle
  
  ' output
  
  '  atan3 = angle (0 =< atan3 <= 2 * pi; radians)
  
  ''''''''''''''''''''''''''''''''''''''''''''''''
  
  local c
  
  if (abs(a) < 1.0e-10) then
    
    atan3 = (1.0 - sgn(b)) * pidiv2
    
    exit function
    
  else
    
    c = (2.0 - sgn(a)) * pidiv2
    
  endif
  
  if (abs(b) < 1.0e-10) then
    
    atan3 = c
    
    exit function
    
  else
    
    atan3 = c + sgn(a) * sgn(b) * (abs(atn(a / b)) - pidiv2)
    
  endif
  
end function
  
  '''''''''''''''''''''''''''
  '''''''''''''''''''''''''''
  
function vecmag(a()) as float
  
  ' vector magnitude function
  
  ' input
  
  '  { a } = column vector ( 3 rows by 1 column )
  
  ' output
  
  '  vecmag = scalar magnitude of vector { a }
  
  vecmag = sqr(a(1) * a(1) + a(2) * a(2) + a(3) * a(3))
  
end function
  
  ''''''''''''''''''''
  ''''''''''''''''''''
  
sub uvector (a(), b())
  
  ' unit vector subroutine
  
  ' input
  
  '  a = column vector (3 rows by 1 column)
  
  ' output
  
  '  b = unit vector (3 rows by 1 column)
  
  '''''''''''''''''''''''''''''''''''''''
  
  local i as integer, amag
  
  amag = vecmag(a())
  
  for i = 1 to 3
    
    if (amag <> 0.0) then
      
      b(i) = a(i) / amag
      
    else
      
      b(i) = 0.0
      
    end if
    
  next i
  
end sub
  
  ''''''''''''''''''''''''''''''
  ''''''''''''''''''''''''''''''
  
function vdot(a(), b()) as float
  
  ' vector dot product function
  
  ' c = { a } dot { b }
  
  ' input
  
  '  n%    = number of rows
  '  { a } = column vector with n rows
  '  { b } = column vector with n rows
  
  ' output
  
  '  vdot = dot product of { a } and { b }
  
  ''''''''''''''''''''''''''''''''''''''''
  
  local c = 0.0
  
  for i% = 1 to 3
    
    c = c + a(i%) * b(i%)
    
  next i%
  
  vdot = c
  
end function
  
  '''''''''''''''''''''''
  '''''''''''''''''''''''
  
sub vcross(a(), b(), c())
  
  ' vector cross product subroutine
  
  ' { c } = { a } x { b }
  
  ' input
  
  '  { a } = vector a ( 3 rows by 1 column )
  '  { b } = vector b ( 3 rows by 1 column )
  
  ' output
  
  '  { c } = { a } x { b } ( 3 rows by 1 column )
  
  c(1) = a(2) * b(3) - a(3) * b(2)
  c(2) = a(3) * b(1) - a(1) * b(3)
  c(3) = a(1) * b(2) - a(2) * b(1)
  
end sub
  
  '''''''''''
  '''''''''''
  
sub read_data
  
  ' read sun and planet data subroutine
  
  '''''''''''''''''''''''''''''''''''''
  
  ' read leap second data
  
  for i% = 1 to 28
    
    read jdleap(i%), leapsec(i%)
    
  next i%
  
  data 2441317.5,  10.0
  data 2441499.5,  11.0
  data 2441683.5,  12.0
  data 2442048.5,  13.0
  data 2442413.5,  14.0
  data 2442778.5,  15.0
  data 2443144.5,  16.0
  data 2443509.5,  17.0
  data 2443874.5,  18.0
  data 2444239.5,  19.0
  data 2444786.5,  20.0
  data 2445151.5,  21.0
  data 2445516.5,  22.0
  data 2446247.5,  23.0
  data 2447161.5,  24.0
  data 2447892.5,  25.0
  data 2448257.5,  26.0
  data 2448804.5,  27.0
  data 2449169.5,  28.0
  data 2449534.5,  29.0
  data 2450083.5,  30.0
  data 2450630.5,  31.0
  data 2451179.5,  32.0
  data 2453736.5,  33.0
  data 2454832.5,  34.0
  DATA 2456109.5,  35.0
  data 2457204.5,  36.0
  data 2457754.5,  37.0
  
end sub
  
  '''''''''''''''
  '''''''''''''''
  
function log10(x)
  
  ' Common logarithm function
  
  ' input
  
  '  x = argument
  
  ' output
  
  '  log10 = common log of x
  
  ''''''''''''''''''''''''''
  
  If (x = 0.0) Then
    
    log10 = -99.0
    
  Else
    
    log10 = 0.434294481903252 * log(x)
    
  End If
  
end function

